import requests
import ipaddress
import json
import logging
import os
import git
import shutil
import sys
import traceback
import argparse
import yaml
from tycho.tycho_utils import TemplateUtils
from tycho.config import Config
from tycho.exceptions import TychoException
from tycho.actions import StartSystemResource, StatusSystemResource, DeleteSystemResource, ModifySystemResource
from kubernetes import client as k8s_client, config as k8s_config

logger = logging.getLogger (__name__)

mem_converter = {
    'M' : lambda v : v * 10 ** 6,
    'G' : lambda v : v * 10 ** 9,
    'T' : lambda v : v * 10 ** 12,
    'P' : lambda v : v * 10 ** 15,
    'E' : lambda v : v * 10 ** 18
}



class TychoService:
    """ Represent a service endpoint. """
    try_minikube = True

    def __init__(self, name, app_id, ip_address, port, sid=None, creation_time=None, username="",utilization={}, conn_string="", workspace_name=""):
        self.name = name
        self.app_id = app_id
        self.ip_address = ip_address
        self.port = port
        self.identifier = sid
        self.creation_time = creation_time
        self.username = username
        self.utilization = utilization
        self.total_util = self.get_utilization (utilization)
        self.conn_string = conn_string
        self.workspace_name = workspace_name
            
    def get_utilization (self, utilization):
        total = {
            "gpu": 0,
            "cpu" : 0,
            "memory" : 0
        }
        for key, val in utilization.items ():
            if 'cpu' in val.keys():
                if 'm' in val['cpu']:
                    total['cpu'] = total['cpu'] + int(val['cpu'].replace ('m', ''))
                else:
                    total['cpu'] = total['cpu'] + int(val['cpu']) * 1000
            if 'nvidia.com/gpu' in val.keys():
                total['gpu'] = total['gpu'] + int(val['nvidia.com/gpu'])
            mem = val['memory'].replace ("i", "")
            """ Run the conversion function designated by the last character of the value on the integer value """
            mem_val = mem_converter[mem[-1]] (int(mem[:-1]))
            total['memory'] = total['memory'] + mem_val
        total['memory'] = f"{total['memory'] / (10 ** 9)}"
        return total

    def __repr__(self):
        b = f"id: {self.identifier} time: {self.creation_time} util: {self.utilization}"
        return f"name: {self.name} ip: {self.ip_address} port: {self.port} user:{self.username} {b}"


class TychoStatus:
    """ A response from a status request. """ 
    def __init__(self, status, result, message):
        self.status = status
        self.services = list(map(lambda v: TychoService (**v), result))
        self.message = message

    def __repr__(self):
        return f"status: {self.status} svcs: {[ str(s) for s in self.services ]} msg: {self.message}"


class TychoSystem:
    """ Represents a running system. """
    def __init__(self, status, result, message):
        self.status = status
        if status == 'error':
            raise TychoException (f"status:{status} result:{result} message:{message}")
        self.name = result['name']
        self.identifier = result['sid']
        self.services = [
            TychoService(name=k, app_id=result['name'], ip_address=v['ip_address'], port=v['port-1'])
            for k, v in result['containers'].items ()
        ]
        self.conn_string = result['conn_string']
        self.message = message

        
class TychoClient:
    """ Python client to Tycho dynamic application deployment API. """

    def __init__(self, url="http://localhost:5000"):
        """ Construct a client.

            :param url: URL of the Tycho API endpoint.
            :type url: string
        """
        self.url = f"{url}/system"
        self.actions = {
            'start': StartSystemResource(),
            'status': StatusSystemResource(),
            'delete': DeleteSystemResource(),
            'modify': ModifySystemResource()
        }
        
    def request (self, service, request):
        """ Send a request to the server. Generic underlayer to all requests. 

            :param service: URL path to the service to invoke.
            :param request: JSON to send to the API endpoint.
        """
        if os.environ.get("REST_API", "false") == "true":
            response = requests.post (f"{self.url}/{service}", json=request)
            result_text = f"HTTP status {response.status_code} received from service: {service}"
            logger.debug (result_text)
            if not response.status_code == 200:
                raise Exception (f"Error: {result_text}")
            result = response.json ()
            logger.debug (json.dumps(result, indent=2))
        else:
            result = self.actions.get(service).post(request)
            logger.debug(f"{result} received from service: {service}")
        return result
    
    def format_name (self, name):
        """ Format a service name to be a valid DNS label.
        
            :param name: Format a name.
        """
        return name.replace (os.sep, '-')

    def parse_env (self, environment):
        return {
            line.split("=", maxsplit=1)[0] : line.split("=", maxsplit=1)[1]
            for line in environment.split ("\n") if '=' in line
        }
        
    def start (self, request):
        """ Start a service. 
        
            The general format of a start request is::

                {
                   "name"   : <name of the system>,
                   "env"    : <JSON dict created from .env environment variables>,
                   "system" : <JSON of a docker-compose yaml>
                } 

            :param request: A request object formatted as above.
            :type request: JSON
            :returns: Returns a TychoSystem object
        """
        response = self.request ("start", request)
        error = response.get('result',{}).get('error',None)
        if error == 'error':
            for e in error:
                logger.error (e)
        return TychoSystem (**response)

    def delete (self, request):
        """ Delete a service. 
            
            Given the GUID of a running service, delete it and all its constituent parts.

            The general format of a delete request is::
        
                {
                   "name" : <GUID>
                }

            :param request: A request formatted as above.
            :type request: JSON
        """
        logger.error (f"-- delete: {json.dumps(request, indent=2)}")
        print (f"-- delete: {json.dumps(request, indent=2)}")
        return self.request ("delete", request)
    
    def status (self, request): 
        """ Get status of running systems.
        
            Get the status of a system by GUID or across systems.

            The format of a request is::
 
                {}

            :param request: Request formatted as above.
            :type request: JSON
        """
        response = self.request ("status", request)
        return TychoStatus (**response)

    def modify(self, request):
        """ Takes in a JSON formatted metadata and specs of a running system.

            Some examples for a request,
            * {"tycho-guid": <guid>, "labels": {"name": <any-name>}, "resources": {"cpu": <cpu>, "memory": <memory>}}
            * {"tycho-guid": <guid>, "labels": {"name": "<any-name>"}}
            * {"tycho-guid": <guid>, "resources": {"cpu": <cpu>, "memory": <memory>}}

            :param request: Request formatted as above
            :type request: json

            :returns: A list of all the patches applied to the system
            :rtype: A list
        """
        response = self.request("modify", request)
        return response

    def up (self, name, system, settings=""):
        """ Bring a service up starting with a docker-compose spec. 
        
            CLI endpoint to start a service on the Tycho compute fabric.::

                tycho up -f path/to/docker-compose.yaml

            :param name: Name of the system.
            :type name: str
            :param system: Docker-compose JSON structure.
            :type system: JSON
            :param settings: The textual contents of a .env file.
            :type settings: str
        """
        services = {}
        for container_name, container in system['services'].items ():
            if 'ports' in container.keys():
                ports = container['ports']
                for port in ports:
                    port_num = int(port.split(':')[1] if ':' in port else port)
                    services[container_name] = {
                        "port" : port_num
                        #"clients" : [ "192.16.1.179" ]
                    }
                    
        request = {
            "name"   : self.format_name (name),
            "principal" : '{"username": "renci"}',
            "serviceaccount" : "default",
            "env"    : self.parse_env (settings),
            "system" : system,
            "services" : services
        }
        logger.debug (f"request: {json.dumps(request, indent=2)}")
        response = self.start (request)
        logger.debug (response)
        if response.status == 'error':
            print (response.message)
        else:
            format_string = '{:<30} {:<35} {:<15} {:<7}'
            print (format_string.format("SERVICE", "GUID", "IP_ADDRESS", "PORT"))
            for service in response.services:
                print (format_string.format (
                    TemplateUtils.trunc (service.name, max_len=28),
                    TemplateUtils.trunc (response.identifier, max_len=33),
                    service.ip_address,
                    service.port))
                break

    def list (self, name, terse=False):
        """ List status of executing systems.

            CLI endpoint to list status of services.::

                tycho status
                tycho status -terse

            :param name: GUID of a service to get status for.
            :type name: str
            :param terse: Print just the GUID for running systems
            :type terse: boolean
        """
        try:
            request = { "name" : self.format_name (name) } if name else {}
            request['username'] = 'renci'
            response = self.status (request)
            logger.debug (response)
            if response.status  == 'success':
                if terse:
                    for service in response.services:
                        print (service.identifier)
                elif len(response.services) == 0:
                    print ('None running')
                else:
                    format_string = '{:<30} {:<35} {:<15} {:<7} {:<1}'
                    print (format_string.format("SYSTEM", "GUID", "IP_ADDRESS", "PORT", "CREATION_TIME"))
                    for service in response.services:
                        print (format_string.format (
                            TemplateUtils.trunc (service.name, max_len=28),
                            TemplateUtils.trunc (service.identifier, max_len=33),
                            service.ip_address,
                            service.port,
                            service.creation_time))
            elif response.status == 'error':
                print (response)
        except Exception as e:
            raise e
        
    def down (self, names):
        """ Bring down a service. 

            CLI endpoint for deleting running systems.::

                tycho down <GUID>

            :param names: GUIDs of systems to delete.
            :type name: str
        """
        try:
            for name in names:
                response = self.delete ({ "name" : self.format_name(name) })
                logger.debug (json.dumps(response,indent=2))
                if response.get('status',None) == 'success':
                    print (f"{name}")
                else:
                    print (json.dumps (response, indent=2))
        except Exception as e:
            traceback.print_exc (e)

    def patch(self, mod_items):
        """
            Modify a running system.
            Takes in a JSON formatted metadata and specs of a running system.

            Some examples for a request,
            * {"tycho-guid": <guid>, "labels": {"name": <any-name>}, "resources": {"cpu": <cpu>, "memory": <memory>}}
            * {"tycho-guid": <guid>, "labels": {"name": "<any-name>"}}
            * {"tycho-guid": <guid>, "resources": {"cpu": <cpu>, "memory": <memory>}}

            CLI endpoint for modifying running systems::
                python client -m <JSON object>

            :param mod_items: Request formatted as above
            :type mod_items: json

            :returns: A list of all the patches applied to the system
            :rtype: A list
        """
        logger.info(f"System specifications and metadata to be modified: {mod_items}")
        try:
            response = self.modify(mod_items)
            logger.debug(json.dumps(response, indent=2))
            return response
        except (AttributeError, Exception) as e:
            logger.exception(f"Error in modifying system. {e}")

            
class TychoClientFactory:
    """ Locate a Tycho API instance in a Kubernetes cluster.

        This is written to work wheter run in-cluster or standalone. If we're running outside of 
        the cluster we use the environment's kubernetes configuration. If we're running 
        insde kubernetes, we use the "in cluster" configuration to locate the configuration.        
    """
    def __init__(self):
        """ Initialize connection to Kubernetes.

            Load the kubernetes configuration in an enviroment appropriate way as described above.

            Then create the K8S API endpoint.
        """
        if os.getenv('KUBERNETES_SERVICE_HOST'):
            logger.debug ("--loading in cluster configuration.")
            k8s_config.load_incluster_config()
        else:
            logger.debug ("--loading kube config, cluster external.")
            k8s_config.load_kube_config()
        api_client = k8s_client.ApiClient()
        self.api = k8s_client.CoreV1Api(api_client)
        
    def get_client (self, name="tycho-api", namespace="default", default_url="http://localhost:5000"):
        """ Locate the client endpoint using the K8s API.

            Locate the Tycho API using the K8S API. We do this by reading services in the
            given namespace with the given name. Then we look for a load balancer IP and port 
            to build a URL. This works for public cloud clusters. With some modification, it
            could work for Minikube but that is a future effort.

            :param name: Name of the Tycho API service in Kubernetes.
            :type name: str
            :param namespace: The namespace the service is deployed to.
            :type namespace: str
        """
        try:
           with open("/var/run/secrets/kubernetes.io/serviceaccount/namespace", "r") as secrets:
               for line in secrets:
                   namespace = line
                   raise Exception
        except Exception as e:
            logger.warning(f"cannot get namespace from file: {e}")

        url = None
        client = None
        try:
            service = self.api.read_namespaced_service(
                name=name,
                namespace=namespace)
            if not service:
                url = default_url
            elif service.status and service.status.load_balancer:
                ip_address = "tycho-api"
                port = service.spec.ports[0].port
                logger.debug (f"located tycho api instance in kube")
                url = f"http://{ip_address}:{port}"
            elif service.spec and len(service.spec.ports) > 0:
                logger.debug ("--looking in minikube for a node port based service.")
                ip = os.popen('minikube ip').read().strip ()
                if len(ip) > 0:
                    try:
                        ipaddress.ip_address (ip)
                        logger.info (f"configuring minikube ip: {ip}")
                        port = service.spec.ports[0].node_port
                        logger.debug (f"located tycho api instance in minikube")
                        url = f"http://{ip}:{port}"
                    except ValueError as e:
                        logger.error ("unable to get minikube ip address")
                        traceback.print_exc (e)
            print(f"URL: {url}")
        except Exception as e:
            url = default_url
            print(f"url: {url}")
            logger.info (f"cannot find {name} in namespace {namespace}")

        logger.info (f"creating tycho client with url: {url}")
        return TychoClient (url=url)


class TychoApps:
    def __init__(self, app):
        self.app = app
        self.repo_url = 'https://github.com/heliumplusdatastage/app-support-prototype.git'
        self.repo_name = 'app-support-prototype'
        self.apps_dir = 'dockstore-yaml-proposals'
        self.source_file = 'docker-compose.yaml'
        self.env = '.env'
        self.metadata = {}

    def deleterepo(self, repo_path):
        shutil.rmtree(repo_path)

    def getmetadata(self):
        base_dir = os.path.dirname(os.path.abspath(__file__))
        try:
            git.Git(base_dir).clone(self.repo_url)
        except Exception as e:
            traceback.print_exc(e)
        repo_path = os.path.join(base_dir, self.repo_name)
        apps_path = os.path.join(repo_path, self.apps_dir)
        for _, dnames, _ in os.walk(apps_path):
            if self.app in dnames:
                system_path = os.path.join(apps_path, self.app, self.source_file)
                env_path = os.path.join(apps_path, self.app, self.env)
                with open(system_path, "r") as stream:
                    system = yaml.safe_load(stream)
                    self.metadata['System'] = system
                if os.path.exists(env_path):
                    with open(env_path, 'r') as stream:
                        settings = stream.read()
                        self.metadata['Settings'] = settings
            break
        self.deleterepo(repo_path)
        return self.metadata


if __name__ == "__main__":
    """ A CLI for Tycho. """
    status_command="@status_command"
    parser = argparse.ArgumentParser(description='Tycho Client')
    parser.add_argument('-u', '--up', help="Launch service.", action='store_true')
    parser.add_argument('-s', '--status', help="Get status of running systems.", nargs='?', const=status_command, default=None)
    parser.add_argument('-d', '--down', help="Delete a running system. Requires a system id.", nargs='*')
    parser.add_argument('-p', '--port', type=int, help="Port to expose.")
    parser.add_argument('-c', '--container', help="Container to run.")
    parser.add_argument('-n', '--name', help="Service name.")
    parser.add_argument('--service', help="Tycho API URL.", default="http://localhost:5000")
    parser.add_argument('--env', help="Env variable", default=None)
    parser.add_argument('--command', help="Container command", default=None)
    parser.add_argument('--settings', help="Environment settings", default=None)
    parser.add_argument('-f', '--file', help="A docker compose (subset) formatted system spec.")
    parser.add_argument('-a', '--app', help="Name of the app. Should be one of the apps available in app-support-prototype repo")
    parser.add_argument('-t', '--trace', help="Trace (debug) logging", action='store_true', default=False)
    parser.add_argument('--terse', help="Keep status short", action='store_true', default=False)
    parser.add_argument('-v', '--volumes', help="Mounts a volume", default=None)
    parser.add_argument('-m', '--modify', help="Modify a running system", default=None)
    args = parser.parse_args ()
                
    """ Honor debug and trace settings. """
    if args.trace:
        logging.basicConfig(level=logging.DEBUG)

    """ Resolve environment settings file as text. """
    settings=""
    if args.settings:
        with open(args.settings, "r") as stream:
            settings = stream.read ()

    name=args.name
    system=None
    if args.app:
        """Name for the app"""
        name = args.app

        """ Apply settings. """
        tychoapps = TychoApps(args.app)
        metadata = tychoapps.getmetadata()

        if 'System' in metadata.keys():
            system = metadata['System']
        if 'Settings' in metadata.keys():
            settings = metadata['Settings']
            print(f"settings: {settings}")

    elif args.file:
        if not args.name:
            """ We've been given a docker-compose.yaml. Come up with a name for the app 
            based on the containing directory if none has been otherwise supplied. """
            if os.sep in args.file:
                args.file = os.path.abspath (args.file)
                name = args.file.split('.')[0] if '.' in args.file else args.file
                name = name.split (os.sep)[-2]
            else:
                name = os.path.basename (os.getcwd ())

        """ Apply settings. """
        env_file = os.path.join (os.path.dirname (args.file), ".env")
        if os.path.exists (env_file):
            with open (env_file, 'r') as stream:
                settings = stream.read ()
                
        with open(args.file, "r") as stream:
            system = yaml.load (stream.read ())
    else:
        """ Generate a docker-compose spec based on the CLI args. """
        name = args.name
        template_utils = TemplateUtils (config=Config())
        template = """
          version: "3"
          services:
            {{args.name}}:
              image: {{args.container}}
              {% if args.command %}
              entrypoint: {{args.command}}
              {% endif %}
              {% if args.port %}
              ports:
                - "{{args.port}}"
              {% endif %}
              {% if args.volumes %}
              volumes:
                - "{{args.volumes}}"
              {% endif %}"""

        system = template_utils.render_text(
            TemplateUtils.apply_environment (settings, template),
            context={ "args" : args })

    client = None
    """ Locate the Tycho API endpoint. Instantiate a client to use the endpoint. """
    if args.service == parser.get_default ("service"):
        """ If the endpoint is the default value, try to discover the endpoint in kube. """
        client_factory = TychoClientFactory ()
        client = client_factory.get_client ()
        if not client:
            """ That didn't work so use the default value. """
            client = TychoClient (url=args.service)
    if not client:
        logger.info (f"creating client directly {args.service}")
        client = TychoClient (url=args.service)

    if args.up:
        client.up (name=name, system=system, settings=settings)
    elif args.down:
        client.down (names=args.down)
    elif args.status:
        if args.status == status_command: # non arg
            client.list (name=None, terse=args.terse)
        else:
            client.list (name=args.status, terse=args.terse)
    elif args.modify:
        mod_items = json.loads(args.modify)
        client.patch(mod_items=mod_items)



import argparse
import logging
import ipaddress
import json
import os
from typing import OrderedDict
import uuid
import yaml
import traceback
from tycho.tycho_utils import TemplateUtils

logger = logging.getLogger (__name__)


class Limits:
    """ Abstraction of resource limits on a container in a system. """
    def __init__(self,
                 cpus=None,
                 gpus=None,
                 memory=None):
        """ Create limits.
            
            :param cpus: Number of CPUs. May be a fraction.
            :type cpus: str
            :param gpus: Number of GPUs. May be a fraction.
            :type gpus: str
            :param memory: Amount of memory 
            :type memory: str
        """
        self.cpus = cpus
        self.gpus = gpus
        #assert (self.gpus).is_integer, "Fractional GPUs not supported"
        self.memory = memory
    def __repr__(self):
        return f"cpus:{self.cpus} gpus:{self.gpus} mem:{self.memory}"


class Volumes:
    def __init__(self, id, containers):
        self.id = id
        self.containers = containers
        self.volumes = []
        self.pvcs = []

    def volume(self, container_name, pvc_name, volume_name, path=None, subpath=None):
        self.volumes.append({"container_name": container_name, "pvc_name": pvc_name, "volume_name": volume_name, "path": path, "subpath": subpath})

    def process_volumes(self):
       for index, container in enumerate(self.containers):
           for index, volume in enumerate(container["volumes"]):
               parts = volume.split(":")
               if parts[0] == "pvc":
                   volume_name = parts[1].split("/")[2:3][0]
                   pvc_name = volume_name if volume_name not in self.pvcs else None
                   self.pvcs.append(volume_name)
                   path = parts[2] if len(parts) is 3 else None
                   subpath = "/".join(parts[1].split("/")[3:]) if len(parts) is 3 else None
                   self.volume(container['name'], pvc_name, volume_name, path, subpath)
               else:
                   logger.debug(f"Volume definition should follow the pattern: pvc://<pvc_name>/<sub-path>:<container-path> or pvc://<sub-path>:<container-path>")
                   raise Exception(f"Wrong Volume definition in Container:{container['name']} and Volume:{volume}")
       return self.volumes

class Probe:
    def __init__(self,cmd=None,delay=None,period=None,threshold=None):
        self.cmd = cmd
        self.delay = delay
        self.period = period
        self.threshold = threshold

class HttpProbe(Probe):
    def __init__(self,delay=None,period=None,threshold=None,httpGet=None):
        Probe.__init__(self,None,delay,period,threshold)
        if httpGet != None:
           self.path = httpGet.get("path","/")
           self.port = httpGet.get("port",80)
           self.httpHeaders = httpGet.get("httpHeaders",None)

class TcpProbe(Probe):
    def __init__(self,delay=None,period=None,threshold=None,tcpSocket=None):
        Probe.__init__(self,None,delay,period,threshold)
        if tcpSocket != None:
           self.port = tcpSocket.get("port",None)

class Container:
    """ Invocation of an image in a specific infastructural context. """
    def __init__(self,
                 name,
                 image,
                 command=None,
                 env=None,
                 identity=None,
                 limits=None,
                 requests=None,
                 ports=[],
                 expose=[],
                 depends_on=None,
                 volumes=None,
                 securityContext=None,
                 liveness_probe=None,
                 readiness_probe=None):
        """ Construct a container.
        
            :param name: Name the running container will be given.
            :param image: Name of the image to use.
            :param command: Text of the command to run.
            :param env: Environment settings
            :type env: dict
            :param identity: UID of the user to run as.
            :type identity: int
            :param limits: Resource limits
            :type limits: dict
            :param requests: Resource requests
            :type limits: dict
            :param ports: Container ports to expose.
            :type ports: list of int
            :param volumes: List of volume mounts <host_path>:<container_path>
            :type volumes: list of str
            :param securityContext: Contains container security context, runAsUser and fsGroup
            :type securityContext: dict
        """
        self.name = name
        self.image = image
        self.identity = identity
        self.limits = Limits(**limits) if isinstance(limits, dict) else limits
        self.requests = Limits(**requests) if isinstance(requests, dict) else requests
        if isinstance(self.limits, list):
            self.limits = self.limits[0] # TODO - not sure why this is a list.
        self.ports = ports
        self.expose = expose
        self.depends_on = depends_on
        self.command = command
        self.env = \
                   list(map(lambda v : list(map(lambda r: str(r), v.split('='))), env)) \
                   if env else []
        self.volumes = volumes
        if liveness_probe != None and liveness_probe.get('httpGet',None) != None:
            self.liveness_probe = HttpProbe(**liveness_probe)
        elif liveness_probe != None and liveness_probe.get('tcpSocket',None) != None:
            self.liveness_probe = TcpProbe(**liveness_probe)
        elif liveness_probe != None:
            self.liveness_probe = Probe(**liveness_probe)
        if readiness_probe != None and readiness_probe.get('httpGet',None) != None:
            self.readiness_probe = HttpProbe(**readiness_probe)
        elif readiness_probe != None and readiness_probe.get('tcpSocket',None) != None:
            self.readiness_probe = TcpProbe(**readiness_probe)
        elif readiness_probe != None:
            self.readiness_probe = Probe(**readiness_probe)
        self.security_context = securityContext

    def __repr__(self):
        return f"name:{self.name} image:{self.image} id:{self.identity} limits:{self.limits}"



class System:
    """ Distributed system of interacting containerized software. """
    def __init__(self, config, name, principal, service_account, conn_string, proxy_rewrite_rule, containers, identifier, services={}):
        """ Construct a new abstract model of a system given a name and set of containers.
        
            Serves as context for the generation of compute cluster specific artifacts.

            :param config: Configuration information.
            :type name: `Config`
            :param name: Name of the system.
            :type name: str
            :param containers: List of container specifications.
            :type containers: list of containers
        """
        self.config = config
        self.identifier = identifier
        self.system_name = name
        self.amb = False
        self.dev_phase = os.getenv('DEV_PHASE', "prod")
        self.name = f"{name}-{self.identifier}"
        assert self.name is not None, "System name is required."
        containers_exist = len(containers) > 0
        none_are_null = not any([ c for c in containers if c == None ])
        assert containers_exist and none_are_null, "System container elements may not be null."
        logger.info(f"=======> Constructing system from containers = {containers}")
        self.containers = list(map(lambda v : Container(**v), containers)) \
                          if isinstance(containers[0], dict) else \
                             containers
        """ Construct a map of services. """
        self.services = {
            service_name : Service(**service_def)
            for service_name, service_def in services.items ()
        }
        for name, service in self.services.items ():
            service.name = f"{name}-{self.identifier}"
            service.name_noid =  name
        self.volumes = Volumes(self.identifier, containers).process_volumes()
        self.source_text = None
        self.system_port = None
        self.ambassador_id = self._get_ambassador_id()
        """ System environment variables """
        self.system_env = dict(principal)
        """ System tags """
        self.username = principal.get("username")
        self.annotations = {}
        self.namespace = "default"
        self.serviceaccount = service_account
        self.runasroot = os.environ.get("RUNASROOT", "true").lower()
        self.conn_string = conn_string
        """PVC flags and other variables for default volumes"""
        self.create_home_dirs = os.environ.get("CREATE_HOME_DIRS", "false").lower()
        self.stdnfs_pvc = os.environ.get("STDNFS_PVC", "stdnfs")
        self.parent_dir = os.environ.get('PARENT_DIR', 'home')
        self.subpath_dir = os.environ.get('SUBPATH_DIR', self.username)
        self.shared_dir = os.environ.get('SHARED_DIR', 'shared')
        """Default UID and GID for the system"""
        security_context = self.config.get('tycho')['compute']['system']['defaults']['securityContext']
        self.Uid = security_context.get('Uid', '1000')
        self.Gid = security_context.get('Gid', '1000')
        """Resources and limits for the init container"""
        self.init_cpus = os.environ.get("INIT_CPUS", "250m")
        self.init_memory = os.environ.get("INIT_MEMORY", "250Mi")
        """Proxy rewrite rule for ambassador service annotations"""
        self.proxy_rewrite_rule = proxy_rewrite_rule

    def _get_ambassador_id(self):
        return os.environ.get("AMBASSADOR_ID", "")
    
    @staticmethod
    def get_identifier():
        return uuid.uuid4().hex

    def _get_init_resources(self):
        resources = self.config.get('tycho')['compute']['system']['defaults']['services']['init']['resources']
        return resources

    def get_namespace(self, namespace="default"):
        try:
           with open("/var/run/secrets/kubernetes.io/serviceaccount/namespace", "r") as secrets:
               for line in secrets:
                   namespace = line
                   break
        except Exception as e:
            logger.warning(f"error getting namespace from file: {e}")
        return namespace

    def requires_network_policy (self):
        return any ([ len(svc.clients) > 0 for name, svc in self.services.items () ])
    
    def render (self, template, context={}):
        """ Supply this system as a context to a template.
        
            :param template: Template 
        """
        final_context = { "system" : self }
        for n, v in context.items ():
            final_context[n] = v
        generator = TemplateUtils (config=self.config)
        template = generator.render (template, context=final_context)
        logger.debug (f"--generated template: {template}")
        return template

    @staticmethod
    def parse (config, name, principal, system, service_account, env={}, services={}):
        """ Construct a system model based on the input request.

            Parses a docker-compose spec into a system specification.
        
            :param name: Name of the system.
            :param system: Parsed docker-compose specification.
            :param env: Dictionary of settings.
            :param services: Service specifications - networking configuration.
        """
        principal = json.loads(principal)
        identifier = System.get_identifier()
        containers = []
        if env != None:
            env['identifier'] = identifier
            env['username'] = principal.get('username',"Unknown")
            system_port = None
            for cname,spec in system.get('services',{}).items():
                 env['system_name'] = cname
                 for p in spec.get('ports', []):
                     if ':' in p: system_port = p.split(':')[1]
                     else: system_port = p
                     break
            if system_port != None: env['system_port'] = system_port
            else: env['system_port'] = 8000
            logger.debug ("applying environment settings.")
            system_template = yaml.dump (system)
            logger.debug (json.dumps(env,indent=2))
            system_rendered = TemplateUtils.render_text(template_text=system_template,context=env)
            logger.debug (f"applied settings:\n {system_rendered}")
            for system_render in system_rendered:
                system = system_render

        """ Model each service. """
        logger.debug (f"compose {system}")
        for cname, spec in system.get('services', {}).items ():
            """ Entrypoint may be a string or an array. Deal with either case."""
            ports = []
            expose = []
            env_all = []
            entrypoint = spec.get ('entrypoint', '')
            """ Adding default volumes to the system containers """
            if spec.get('volumes') == None:
                spec.update({'volumes': []})
            rep = {
                'stdnfs_pvc': os.environ.get('STDNFS_PVC', 'stdnfs'), 
                'username': principal.get("username"),
                'parent_dir': os.environ.get('PARENT_DIR', 'home'),
                'subpath_dir': os.environ.get('SUBPATH_DIR', principal.get("username")),
                'shared_dir': os.environ.get('SHARED_DIR', 'shared'),
            }
            if os.environ.get("DEV_PHASE", "prod") != "test":
                try:
                    for volume in config.get('tycho')['compute']['system']['volumes']:
                        createHomeDirs = os.environ.get('CREATE_HOME_DIRS', "true")
                        volSplit = volume.split(":")
                        if createHomeDirs == "false" and ("username" in volume or "shared_dir" in volSplit[1]):
                            continue
                        if createHomeDirs == "true" and ("shared_dir" not in volSplit[1] and "subpath_dir" not in volSplit[2]):
                            continue
                        for k, v in rep.items():
                            volume = volume.replace(k, v)
                        spec.get('volumes', []).append(volume)
                except Exception as e:
                    logger.info("No volumes specified in the configuration.")
            """ Adding entrypoint to container if exists """
            if isinstance(entrypoint, str):
                entrypoint = entrypoint.split ()
            for p in spec.get('ports', []):
              if ':' in p:
                ports.append({
                  'containerPort': p.split(':')[1]
                })
              else:
                ports.append({
                  'containerPort': p
                })
            for e in spec.get('expose', []):
              expose.append({
                'containerPort': e
              })
            """Parsing env variables"""
            env_from_spec = (spec.get('env', []) or spec.get('environment', []))
            env_from_registry = [f"{ev}={os.environ.get('STDNFS_PVC')}" if '$STDNFS' in env[ev] else f"{ev}={env[ev]}" for ev in env]
            env_all = env_from_spec + env_from_registry
            if spec.get("ext",None) != None and spec.get("ext").get("kube",None) != None:
                liveness_probe = spec["ext"]["kube"].get('livenessProbe',None)
                readiness_probe = spec["ext"]["kube"].get('readinessProbe',None)
                if isinstance(liveness_probe,str) and liveness_probe == "none": liveness_probe = None
                if isinstance(readiness_probe,str) and readiness_probe == "none": readiness_probe = None
            else:
                liveness_probe = None
                readiness_probe = None
            containers.append({
                "name": cname,
                "image": spec['image'],
                "command": entrypoint,
                "env": env_all,
                "limits": spec.get('deploy',{}).get('resources',{}).get('limits',{}),
                "requests": spec.get('deploy',{}).get('resources',{}).get('reservations',{}),
                "ports": ports,
                "expose": expose,
                "depends_on": spec.get("depends_on", []),
                "volumes": [v for v in spec.get("volumes", [])],
                "securityContext":  spec.get("securityContext", {}),
                "liveness_probe": liveness_probe,
                "readiness_probe": readiness_probe
            })
        system_specification = {
            "config": config,
            "name": name,
            "principal": principal,
            "service_account": service_account,
            "conn_string": spec.get("conn_string", ""),
            "proxy_rewrite_rule": spec.get("proxy_rewrite_rule", False),
            "containers": containers,
            "identifier": identifier,
            "services": services
        }
        logger.debug (f"parsed-system: {json.dumps(system_specification, indent=2)}")
        system = System(**system_specification)
        system.source_text = yaml.dump (system)
        return system

    def __repr__(self):
        return f"name:{self.name} containers:{self.containers}"


class ModifySystem:
    """
       This is a class representation of a system's metadata and specs that needs to be modified.

       :param config: A default config for Tycho
       :type config: A dict
       :param guid: A unique guid to a system/deployment
       :type guid: The UUID as a 32-character hexadecimal string
       :param labels: A dictionary of labels that are applied to deployments
       :type labels: A dictionary
       :param resources: A dictionary containing cpu and memory as keys
       :type resources: A dictionary
       :param containers: A list of containers that are applied to resources
       :type containers: A list of Kubernetes V1Container objects, optional
    """
    def __init__(self, config, patch, guid, labels, resources):
        """
           A constructor method to ModifySystem
        """
        self.config = config
        self.patch = patch
        self.guid = guid
        self.labels = labels
        self.resources = resources
        self.containers = []

    @staticmethod
    def parse_modify(config, guid, labels, cpu, memory):
        """
           Returns an instance of :class:`tycho.model.ModifySystem` class

           :returns: An instance of ModifySystem class
           :rtype: A class object
        """

        resources = {}
        if cpu is not None:
            resources.update({"cpu": cpu})
        if memory is not None:
            resources.update({"memory": memory})

        if len(resources) > 0 or len(labels) > 0:
            patch = True
        else:
            patch = False

        modify_system = ModifySystem(
            config,
            patch,
            guid,
            labels,
            resources,
        )
        return modify_system

    def __repr__(self):
        return f"name: {self.guid} labels: {self.labels} resources: {self.resources}"


class Service:
    """ Model network connectivity rules to the system. """
    def __init__(self, port=None, clients=[]):
        """ Construct a service object modeling network connectivity to a system. """
        self.port = port
        self.clients = list(map(lambda v: str(ipaddress.ip_network (v)), clients))
        self.name = None
        self.name_noid = None
        
    def __repr__(self):
        return json.dumps (
            f"service: {json.dumps({'port':self.port,'clients':self.clients}, indent=2)}")



import argparse
import json
import logging
import os
import shutil
import subprocess
import sys
import traceback
import yaml
import copy
import base64
from time import sleep
from kubernetes import client as k8s_client, config as k8s_config
from tycho.compute import Compute
from tycho.exceptions import DeleteException
from tycho.exceptions import StartException
from tycho.exceptions import TychoException
from tycho.exceptions import ModifyException
from tycho.model import System
from tycho.tycho_utils import TemplateUtils
import kubernetes.client
from kubernetes.client.rest import ApiException

logger = logging.getLogger (__name__)

port_forwards = {}


class KubernetesCompute(Compute):
    """ A Kubernetes orchestrator implementation.

        Tested with Minikube and Google Kubernetes Engine.
    """

    def __init__(self, config):
        """ Initialize connection to Kubernetes. 
        
            Connects to Kubernetes configuration using an environment appropriate method.
        """
        super(KubernetesCompute, self).__init__()
        self.config = config
        if os.getenv('KUBERNETES_SERVICE_HOST'):
            """ We're running inside K8S. Load the config appropriately. """
            k8s_config.load_incluster_config()
        else:
            """ We're running outside of K8S. Load the config. """
            k8s_config.load_kube_config()
        api_client = k8s_client.ApiClient()
        self.api = k8s_client.CoreV1Api(api_client)
        self.rbac_api = k8s_client.RbacAuthorizationV1Api(api_client)
#        self.extensions_api = k8s_client.ExtensionsV1beta1Api(api_client) 
        self.extensions_api = k8s_client.AppsV1Api(api_client)
        self.networking_api = k8s_client.NetworkingV1Api(api_client)
        self.try_minikube = True
        self.namespace = self.get_namespace (
            namespace=os.environ.get("NAMESPACE", self.get_namespace ()))
        logger.debug (f"-- using namespace: {self.namespace}")
        
    def get_namespace(self, namespace="default"):
        try:
            with open("/var/run/secrets/kubernetes.io/serviceaccount/namespace", "r") as secrets:
                for line in secrets:
                    namespace = line
                    break
        except Exception as e:
            logger.debug(f"-- downward api namespace lookup failed.")
        return namespace

    def check_volumes(self, volumes, namespace):
        try:
            volumesNA = []
            api_response = self.api.list_namespaced_persistent_volume_claim(namespace=namespace)
            for index, volume in enumerate(volumes):
                notExists = True
                if volume["volume_name"] != "stdnfs":
                    for item in api_response.items:
                        if item.metadata.name != volume["volume_name"]:
                            continue
                        else:
                            notExists = False
                            logger.info(f"PVC {volume['volume_name']} exists.")
                            break
                if notExists and volume["volume_name"] != 'stdnfs':
                    volumesNA.append(index)
                    #raise Exception(f"Cannot create system. PVC {volume['pvc_name']} does not exist. Create it.")
            return volumesNA 
        except Exception as e:
            logger.debug(f"Raising persistent volume claim exception. {e}")
            #raise

    def is_ambassador_context(self, namespace):
        try:
            api_response = self.api.list_namespaced_service(field_selector="metadata.name=ambassador", namespace=namespace)
            if len(api_response.items) > 0:
                return True
            else:
                return False
        except ApiException as e:
            logger.info(f"Amabassador is not configured.")

    def start (self, system, namespace="default"):
        """ Start an abstractly described distributed system on the cluster.
            Generate each required K8s artifact and wire them together. Currently 
            explicitly modeled elements include Deployment, PersistentVolume, 
            PersistentVolumeClaim, and Service components.

            :param system: An abstract system model.
            :type system: :class:`.System`
            :param namespace: Namespace to run the system in.
            :type namespace: str
        """
        namespace = self.namespace #system.get_namespace()
        try:
            """ Check volumes and remove them from the system. """
            volumesNA = self.check_volumes(system.volumes, namespace)
            systemVolumesCopy = []
            for index, value in enumerate(system.volumes):
                if index not in volumesNA:
                    systemVolumesCopy.append(value)
            system.volumes = systemVolumesCopy
            """ Check the status of ambassador """
            amb_status = self.is_ambassador_context(namespace)
            if amb_status:
                system.amb = True
            #api_response = self.api.list_namespace()
            #notExists = True
            #for item in api_response.items:
            #    link = item.metadata.self_link
            #    app_ns = link.split("/")[-1]
            #    if app_ns == system.system_name:
            #        notExists = False
            #        logger.info(f"Namespace {system.system_name} exists. Skipping create.")
            #        break
            #if notExists:
            #    ns_manifests = system.render(template="namespace.yaml")
            #    for ns_manifest in ns_manifests:
            #        logger.info(f"Namespace {system.system_name} created.")
            #        api_response = self.api.create_namespace(body=ns_manifest)

            try:
                api_response = self.api.list_namespaced_secret(namespace=namespace)
                for item in api_response.items:
                    if item.metadata.name == f"{system.system_name}-env":
                        for key, value in item.data.items():
                            value = str(base64.b64decode(value), 'utf-8')
                            for container in system.containers:
                                container.env.append([key, value])
                        break
            except ApiException as e:
                logger.debug(f"App requires {system.system_name}-env configmap with envs: {e}")
                ## TODO: Swallows exception.
                
            try:
                for container in system.containers:
                    if container.name == system.system_name:
                        for port in container.ports:
                            system.system_port = port['containerPort']
                            break
                    break
            except Exception as e:
                traceback.print_exc (e)
                exc_type, exc_value, exc_traceback = sys.exc_info()
                text = traceback.format_exception(
                    exc_type, exc_value, exc_traceback)
                raise TychoException (
                    message=f"Failed to get system port:",
                    details=text)
                ## TODO: Why not catch at the end?
                
            """ Turn an abstract system model into a cluster specific representation. """
            pod_manifests = system.render ("pod.yaml")
            #return {}
            #""" Render a persistent volume claim. """
            #pvc_manifests = system.render(template="pvc.yaml")
            #""" Create persistent volume claims. """
            #for pvc_manifest in pvc_manifests:
            #    if pvc_manifest["metadata"]["name"] != "nfs":
            #        response = self.api.create_namespaced_persistent_volume_claim(
            #            namespace=namespace,
            #            body=pvc_manifest)
            #""" Render persistent volumes. """
            #pv_manifests = system.render(template="pv.yaml")
            #""" Create the persistent volumes. """
            #for pv_manifest in pv_manifests:
            #    response = self.api.create_persistent_volume(
            #        body=pv_manifest)

            """ Create a deployment for the pod. """
            for pod_manifest in pod_manifests:
                deployment,create_deployment_api_response = self.pod_to_deployment (
                    name=system.name,
                    username=system.username,
                    identifier=system.identifier,
                    template=pod_manifest,
                    namespace=namespace)

            """ Create a network policy if appropriate. """
            if system.requires_network_policy ():
                logger.debug ("creating network policy")
                network_policy_manifests = system.render (
                    template="policy/tycho-default-netpolicy.yaml")
                for network_policy_manifest in network_policy_manifests:
                    logger.debug (f"applying network policy: {network_policy_manifest}")
                    network_policy = self.networking_api.create_namespaced_network_policy (
                        body=network_policy_manifest,
                        namespace=namespace)

            """ Create service endpoints. """
            container_map = {}
            counter = 0
            for container in system.containers:
                """ Determine if a service is configured for this container. """
                service = system.services.get (container.name, None)
                if service:
                    logger.debug (f"generating service for container {container.name}")
                    service_manifests = system.render (
                        template = "service.yaml",
                        context = { "service" : service, "create_deployment_api_response":create_deployment_api_response }
                    )
                    for service_manifest in service_manifests:
                        logger.debug (f"-- creating service for container {container.name}")                        
                        response = self.api.create_namespaced_service(
                            body=service_manifest,
                            namespace=namespace)

                        ip_address = None
                        if not system.amb:
                            ip_address = self.get_service_ip_address (response)

                        """ Return generated node ports to caller. """
                        for port in response.spec.ports:
                            container_map[container.name] = {
                                "ip_address" : ip_address,
                                port.name    : port.node_port
                            }
                            break
            result = {
                'name'       : system.name,
                'sid'        : system.identifier,
                'containers' : container_map,
                'conn_string': system.conn_string
            }
        
        except Exception as e:
            self.delete (system.name)
            exc_type, exc_value, exc_traceback = sys.exc_info()
            text = traceback.format_exception(
                exc_type, exc_value, exc_traceback)
            raise StartException (
                message=f"Unable to start system: {system.name}",
                details=text)

        logger.info (f"result of the app launch: {json.dumps(result,indent=2)}")
        return result

    def get_service_ip_address (self, service_metadata):
        """ Get the IP address for a service. On a system with a load balancer
            that will be in the service status' load balancer section. On minikube,
            we use the minikube IP address which is in the system config. 


            :param service: Service metadata.
            :returns: ip_address IP Address of the service.
            """
        ip_address = None if os.environ.get("DEV_PHASE", "prod") != "test" else "127.0.0.1"
        try:
            app_id = service_metadata.metadata.labels["tycho-app"]
            logger.info (f"-================================> *** {app_id}")
            if not app_id in port_forwards:
                port_forwards[app_id] = app_id #process.pid
                sleep (3)
                logger.debug (f"--------------> {service_metadata.spec.ports}")
                port = service_metadata.spec.ports[0].port
                node_port = service_metadata.spec.ports[0].node_port
                if node_port is None:
                    node_port = service_metadata.spec.ports[0].target_port
                exe = shutil.which ('kubectl')
                command = f"{exe} port-forward --pod-running-timeout=3m0s deployment/{app_id} {node_port}:{port}"
                logger.debug (f"-- port-forward: {command}")
                process = subprocess.Popen (command,
                                            shell=True,
                                            stderr=subprocess.STDOUT)
                """ process dies when the other end disconnects so no need to clean up in delete. """
            #ip_address = "127.0.0.1"
        except Exception as e:
            traceback.print_exc ()
        logger.debug (f"service {service_metadata.metadata.name} ingress ip: {ip_address}")
        '''
        if not ip_address:
            if self.try_minikube:
                try:
                    ip_address = os.popen ("minikube ip").read ().strip ()
                except Exception as e:
                    self.try_minikube = False
                    # otherwise not an error, just means we're not using minikube.
        '''
        return ip_address

    def pod_to_deployment (self, name, username, identifier, template, namespace="default"):
        """ Create a deployment specification based on a pod template.
            
            :param name: Name of the system.
            :type name: str
            :param template: Relative path to the template to use.
            :type template: str
            :param identifier: Unique key to this system.
            :type identifier: str
            :param namepsace: Namespace to run the pod in.
            :type namespace: str
        """
        namespace = self.namespace #self.get_namespace()
#        deployment_spec = k8s_client.ExtensionsV1beta1DeploymentSpec(
        deployment_spec = k8s_client.V1DeploymentSpec(
            replicas=1,
            template=template,
            selector=k8s_client.V1LabelSelector (
                match_labels = {
                    "tycho-guid" : identifier,
                    "username"   : username
                }))
        
        """ Instantiate the deployment object """
        logger.debug (f"creating deployment specification {template}")
        deployment = k8s_client.V1Deployment(
            api_version="apps/v1",
            kind="Deployment",
            metadata=k8s_client.V1ObjectMeta(
                name=name,
                labels={
                    "tycho-guid" : identifier,
                    "executor" : "tycho",
                    "username" : username
                }),
            spec=deployment_spec)

        """ Create the deployment. """
        logger.debug (f"applying deployment {template}")
        api_response = self.extensions_api.create_namespaced_deployment(
            body=deployment,
            namespace=namespace)
        logger.debug (f"deployment created. status={api_response.status}")
        return deployment,api_response

    def delete (self, name, namespace="default"):
        """ Delete the deployment. 
                
            :param name: GUID of the system to delete.
            :type name: str
            :param namespace: Namespace the system runs in.
            :type namespace: str
        """
        namespace = self.namespace #self.get_namespace()
        try: 
            """ Treat everything with a namespace parameterized collections based delete 
            operator the same. """
            finalizers = {
                "deployment"  : self.extensions_api.delete_collection_namespaced_deployment,
                "replica_set" : self.extensions_api.delete_collection_namespaced_replica_set,
                "pod"         : self.api.delete_collection_namespaced_pod,
                "persistentvolumeclaim" : self.api.delete_collection_namespaced_persistent_volume_claim,
                #"networkpolicy" : self.networking_api.delete_collection_namespaced_network_policy
            }
            for object_type, finalizer in finalizers.items ():
                logger.debug (f" --deleting {object_type} elements of {name} in namespace {namespace}")
                response = finalizer (
                    label_selector=f"tycho-guid={name}",
                    namespace=namespace)
            
        except ApiException as e:
            traceback.print_exc (e)
            exc_type, exc_value, exc_traceback = sys.exc_info()
            text = traceback.format_exception(
                exc_type, exc_value, exc_traceback)
            raise DeleteException (
                message=f"Failed to delete system: {name}",
                details=text)
        return {
        }
    
    def status (self, name=None, username=None, namespace="default"):
        """ Get status.
            Without a name, this will get status for all running systems.
            With a name, it will get status for the specified system.

            :param name: GUID of a system to get status for.
            :type name: str
            :param namespace: Namespace the system runs in.
            :type namespace: str
        """
        namespace = self.namespace
        result = []

        """ Find all our generated deployments. """
        label = f"tycho-guid={name}" if name else f"executor=tycho"
        if username:
            label = f"username={username}" if username else f"executor=tycho"
        logger.debug (f"-- status label: {label}")
        response = self.extensions_api.list_namespaced_deployment (
            namespace,
            label_selector=label)

        if response is not None:
            for item in response.items:

                """ Collect pod metrics for this deployment. """
                pod_resources = {
                    container.name : container.resources.limits
                    for container in item.spec.template.spec.containers
                }
                logger.debug(f"-- pod-resources {pod_resources}")

                item_guid = item.metadata.labels.get("tycho-guid", None)
                item_username = item.metadata.labels.get("username", None)

                """ Get the creation timestamp"""
                c_time = item.metadata.creation_timestamp
                time = f"{c_time.month}-{c_time.day}-{c_time.year} {c_time.hour}:{c_time.minute}:{c_time.second}"

                """ Get the workspace name of the pod """
                workspace_name = item.spec.template.metadata.labels.get("app-name", "")

                """ Temporary variables so rest of the code doesn't break elsewhere. """
                ip_address = "127.0.0.1"
                port = 80

                result.append(
                    {
                        "name": item.metadata.name,
                        "app_id": item.spec.template.metadata.labels.get('original-app-name', None),
                        "sid": item_guid,
                        "ip_address": ip_address,
                        "port": str(port),
                        "creation_time": time,
                        "username": item_username,
                        "utilization": pod_resources,
                        "workspace_name": workspace_name
                    }
                )

        return result

    def modify(self, system_modify):
        """
           Returns a list of all patches,

               * metadata labels - Applied to each deployment resource including the pods managed by it.
               * container resources - Are applied to each container in the pod managed by a deployment.

           Takes in a handle :class:`tycho.model.ModifySystem` with the following instance variables,

               * config - A default config for Tycho.
               * guid - A unique guid to a system/deployment.
               * labels - A dictionary of labels.
               * resources - A dictionary containing cpu and memory as keys.
               * containers - A list of containers the resources are applied to.

           :param system_modify: Spec and Metadata object
           :type system_modify: class ModifySystem

           :returns: A list of patches applied
           :rtype: A list

        """
        namespace = self.namespace
        patches_applied = list()
        try:
            api_response = self.extensions_api.list_namespaced_deployment(
                label_selector=f"tycho-guid={system_modify.guid}",
                namespace=namespace).items

            if len(api_response) == 0:
                raise Exception("No deployments found. Specify a valid GUID. Format {'guid': '<name>'}.")

            for deployment in api_response:

                system_modify.containers = list()
                # Need this step to get a comprehensive list of containers if it's multi container pod.
                # Later for patching would need a merge key "name" and corresponding image.
                containers = deployment.spec.template.spec.containers
                system_modify.containers = containers

                generator = TemplateUtils(config=system_modify.config)
                templates = list(generator.render("patch.yaml", context={"system_modify": system_modify}))
                patch_template = templates[0] if len(templates) > 0 else {}
                patches_applied.append(patch_template)

                _ = self.extensions_api.patch_namespaced_deployment(
                        name=deployment.metadata.name,
                        namespace=namespace,
                        body=patch_template
                    )

            return {
                "patches": patches_applied
            }

        except (IndexError, ApiException, Exception) as e:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            text = traceback.format_exception(exc_type, exc_value, exc_traceback)
            raise ModifyException(
                message=f"Failed to modify system: {system_modify.guid}",
                details=text
            )







'''
our-pvc:
  - dicom images                    ro-sidecar
  - nfsrods for napari and imagej   irods
  - deepgtex                        rwm

Given:
==============================================================
docker-compose.yaml:
   ...
   volumes:
     pvc://deepgtex:/deepgtex:rw                    RW!
     pvc://nfsrods:/nfsrods:rw                      RW?     
--------------------------------------------------------------    
every container
*   pvc://stdnfs/home/${username} -> /home/${username}   RW
*   pvc://stdnfs/data             -> /shared             R
    sidecar?                      -> /data               R

options
   user provides a pvc  -> use that as pvc://stdnfs

------

cluster A:
  my-pvc is a RWM pvc.
  bin/tycho api --docker --pvc my-pvc

cluster B:
  pvc-2 is a RWM pvc
  bin/tycho api --docker --pvc pvc-2

------

clusterA:
  my-rwm-storage-class
  bin/tycho api --docker --rwm-sc my-rwm-storage-class
'''



import json
import logging
import os
import random
import shutil
import glob
import subprocess
import sys
import threading
import traceback
import yaml
from tycho.compute import Compute
from tycho.model import System
from tycho.tycho_utils import TemplateUtils
from tycho.exceptions import DeleteException
from tycho.exceptions import StartException
from compose.cli.main import TopLevelCommand, project_from_options

logger = logging.getLogger (__name__)

class DockerComposeThread(threading.Thread):
    """ Run Docker-Compose in a thread and communicate via subprocess. """

    def __init__(self, system, port, configured, app_root):
        """ Invoke thread init and connect the system. """
        threading.Thread.__init__(self)
        self.system = system
        self.port = port
        self.container_map = {}
        self.configured = configured
        self.app_root = app_root
        
    def run (self):
        """ Execute the system. """
        logger.debug (f"creating compose app: {self.system.identifier}")
        os.makedirs (self.app_root)
        #docker_compose_file = os.path.join (self.app_root, f"docker-compose.yaml")
        docker_compose_file = os.path.join (self.app_root, f"{self.system.name}.yaml")
        env_file = os.path.join (self.app_root, ".env")

        """ For now, write literal input text. TODO, generate to incoporate policy. """
        with open (docker_compose_file, 'w') as stream:
            stream.write (self.system.source_text)
        env = f"""HOST_PORT={self.port}\nLOCAL_STORE=./\n"""
        print (f"--env----------> {env}")
        with open (env_file, 'w') as stream:
            stream.write (env)

        """ Find and return ports for each container. """
        config = yaml.load (TemplateUtils.apply_environment (
            env,
            self.system.source_text))
        logger.debug (f"Building conainer map for system {self.system.name}")
        for c_name, c_config in config.get ('services', {}).items ():
            print (f"--cname:{c_name} c_config:{c_config}")
            self.container_map[c_name] = {
                f"{c_name}-{i}" : port.split(':')[0] if ':' in port else port
                for i, port in enumerate(c_config.get('ports', []))
            }
            print (f"-- container map {self.container_map}")

        self.configured.set ()
        
        # Run docker-compose in the directory.
        logger.debug (f"Garbage collecting unused docker networks...")
        p = subprocess.Popen(
            [ "docker", "network", "prune", "--force" ],
            stdout=subprocess.PIPE,
            cwd=self.app_root)

        logger.debug (f"Running system {self.system.name} in docker-compose")
        command = f"docker-compose --project-name {self.system.name} -f {self.system.name}.yaml up --detach"
        print (command)
        p = subprocess.Popen(
            command.split (),
            stdout=subprocess.PIPE,
            cwd=self.app_root)
        
class DockerComposeCompute(Compute):
    def __init__(self, config):
        self.config = config
        self.app_root_base = "apps"
    def start (self, system, namespace="default"):
        import subprocess
        """ Generate a globally unique identifier for the application. All associated 
        objects will share this identifier. """

        app_root = os.path.join (self.app_root_base, system.identifier)
        
        """ Generate a unique port for the system. Needs to be generalized to multi-container 
        while somehow preserving locally meaningful port names."""
        port = random.randint (30000, 40000)
        configured = threading.Event()
        docker_compose_thread = DockerComposeThread (system, port, configured, app_root)
        docker_compose_thread.start ()
        """ Wait for the thread to configure the app to run.
        If this takes longer than five seconds, the thread has probably errored. Continue. """
        configured.wait (5) 
        return {
            'name' : system.name,
            'sid' : system.identifier,
            'containers' : docker_compose_thread.container_map
        }

    def status (self, name, namespace="default"):
        """ Report status of running systems. """
        print (os.getcwd ())
        apps = [ guid for guid in os.listdir(self.app_root_base)
                 if os.path.isdir(os.path.join(self.app_root_base, guid)) ]
        result = []
        cur_dir = os.getcwd ()
        for app in apps:
            app_root = os.path.join (self.app_root_base, app)

            command = f"docker-compose --project-name {app} ps".split ()
            logger.debug (f"--command: {command}")
            p = subprocess.Popen(
                command,
                stdout=subprocess.PIPE,
                cwd=app_root)

            p.stdout.readline ()
            for line in p.stdout:
                print (line)

            result.append ({
                "name" : "--",
                "sid"  : app,
                #"ip"   : None, #"--", #ip_address,
                "port" : "--"
            })
        return result

    def delete (self, name, namespace="default"):
        """ Delete the running process and persistent artifacts. """
        app_root = os.path.join (self.app_root_base, name)

        pattern = os.path.join (app_root, f"*{name}*.yaml")
        print (pattern)
        docker_compose_file = glob.glob (pattern)
        print (docker_compose_file)
        docker_compose_file = os.path.basename (docker_compose_file[0])
        project_name = docker_compose_file.replace (".yaml", "")
        print (f"--project name: {project_name}")
        
        command = f"docker-compose --project-name {project_name} -f {docker_compose_file} down".split ()
        """ Wait for shutdown to complete. """
        p = subprocess.check_call(
            command,
            stdout=subprocess.PIPE,
            cwd=app_root)
        
        """ Delete the app subtree. """
        shutil.rmtree (app_root)



import json
import logging
from tycho.kube import KubernetesCompute
from tycho.dockerc import DockerComposeCompute

logger = logging.getLogger (__name__)

config = {
    "backplane" : "kubernetes"
}
config_factory = {
    "kubernetes"     : KubernetesCompute,
    "docker-compose" : DockerComposeCompute
}
supported_backplanes = config_factory.keys ()

class ComputeFactory:
    
    @staticmethod
    def is_valid_backplane (backplane):
        return backplane in supported_backplanes
    
    @staticmethod
    def set_backplane (backplane):
        config['backplane'] = backplane
        
    @staticmethod
    def create_compute (config):
        backplane = config['tycho']['backplane']
        return config_factory[backplane](config=config)



class TychoException(Exception):
    def __init__(self, message, details=""):
        super().__init__(message)
        self.details = details


class StartException(Exception):
    def __init__(self, message, details=""):
        super().__init__(message, details)


class DeleteException(Exception):
    def __init__(self, message, details=""):
        super().__init__(message, details)


class ContextException(TychoException):
    def __init__(self, message, details=""):
        super().__init__(message, details)


class ModifyException(Exception):
    def __init__(self, message, details=""):
        super().__init__(message, details)


class TemplateException(Exception):
    def __init__(self, message, details=""):
        super().__init__(message)
        self.details = details


class AppRegistryRenderException(Exception):
    def __init__(self, message, details=""):
        super().__init__(message, details)




import ipaddress
import json
import logging
import os
import yaml
import traceback
import re
from tycho.tycho_utils import Resource

logger = logging.getLogger (__name__)

class Config(dict): 
    """ Handle configuration for the system. """
    def __init__(self, config="conf/tycho.yaml"):
        """ Load the system configuration. """
        if isinstance(config, str):
            config_path = Resource.get_resource_path (config)
            logger.debug (f"loading config: {config_path}")
            with open(config_path, 'r') as f:
                self.conf = yaml.safe_load (f)
        elif isinstance(config, dict):
            self.conf = config
        else:
            raise ValueError

        """ Determine if we're on minikube. If so, share its ip address via
        the config. """
        #logger.debug (f"loaded config: {json.dumps(self.conf,indent=2)}")
        if 'TYCHO_ON_MINIKUBE' in os.environ:
            ip = os.popen('minikube ip').read().strip ()
            if len(ip) > 0:
                try:
                    ipaddress.ip_address (ip)
                    logger.info (f"configuring minikube ip: {ip}")
                    self.conf['tycho']['compute']['platform']['kube']['ip'] = ip
                except ValueError as e:
                    logger.error ("unable to get minikube ip address")
                    traceback.print_exc (e)

    def __setitem__(self, key, val):
        self.conf.__setitem__(key, val)
    def __str__(self):
        return self.conf.__str__()
    def __getitem__(self, key):
        return self.conf.__getitem__(key)
    def get (self, key, default=None):
        return self.conf.get(key, default)



import json
import logging
import os
import requests
import requests_cache
import traceback
import uuid
import yaml
import copy
from deepmerge import Merger
from requests_cache import CachedSession
from string import Template
from jinja2 import Template as jinja2Template
from tycho.client import TychoClientFactory, TychoStatus, TychoSystem, TychoClient
from tycho.exceptions import ContextException

logger = logging.getLogger (__name__)

mixin_merge = Merger(
    [ (list,["override"]), (dict,["merge"]), (set,["union"]) ],
    ["override"],
    ["override"]
)

class Principal:
    """ Abstract representation of a system identity. """
    def __init__(self, username, a_token=None, r_token=None):
        self.username=username
        self.access_token=a_token
        self.refresh_token=r_token
        
class TychoContext:
    """
    Load, understand, and use the app registry.

    The app registry is a declarative metadata repository outlining apps available to 
    the platform. Its YAML definition structure provides 
      * Basic metadata about the registry itself including identifier, version, name, etc.
      * A list of repositories or locations apps might reference for further metadata.
      * 
    """
    
    """ https://github.com/heliumdatacommons/CommonsShare_AppStore/blob/master/CS_AppsStore/cloudtop_imagej/deployment.py """
    def __init__(self, registry_config="app-registry.yaml", app_defaults_config="app-defaults.yaml",product="common", stub=False):
        self.registry = self._get_config(registry_config)
        self.app_defaults = self._get_config(app_defaults_config)
        logger.info("defaults = ")
        logger.info(self.app_defaults)
        """ Uncomment this and related lines when this code goes live,. 
        Use a timeout on the API so the unit tests are not slowed down. """
        if not os.environ.get ('DEV_PHASE') == 'stub':
            self.client=TychoClient(url=os.environ.get('TYCHO_URL', "http://localhost:5000"))
        self.product = product
        self.apps = self._grok ()
        self.http_session = CachedSession (cache_name='tycho-registry')

    def _get_config(self, file_name):
        """ Load the registry metadata. """
        config = {}
        """ Load it from the Tycho conf directory for now. Perhaps more dynamic in the future. """
        config_path = os.path.join (
            os.path.dirname (__file__),
            "conf",
            file_name)
        with open(config_path, 'r') as stream:
            config = yaml.safe_load (stream)
        return config

    def add_conf_impl(self, apps, context):
        for key, value in context.items():
            if key in apps.keys():
                apps[key] = {**apps[key], **value}
                self.add_conf_impl(apps, value)
        return apps

    def inherit (self, contexts, context, apps={}):
        for base in context.get ("extends", []):
            self.inherit (contexts, contexts[base], apps)
        apps.update (copy.deepcopy(context.get ("apps", {})))
        return apps
    
    def mixin_defaults(self,apps):
        for app in apps:
            apps[app] = mixin_merge.merge(copy.deepcopy(apps[app]),copy.deepcopy(self.app_defaults))
    
    def mixin(self,contexts,context,apps):
        for base in context.get ("extends", []):
            self.mixin(contexts,contexts[base],apps)
        for mixer in context.get("mixin", []):
            for app in apps:
                if contexts.get(mixer,None) != None and contexts[mixer].get("apps",None) != None and contexts[mixer]["apps"].get(app,None) != None:
                    logger.info("mixing " + app)
                    apps[app] = mixin_merge.merge(copy.deepcopy(apps[app]),copy.deepcopy(contexts[mixer]["apps"].get(app))) 
        return apps

    def _grok (self):
        """ Compile the registry, resolving text substituations, etc. """
        apps = {}
        contexts = self.registry.get ('contexts', {})
        if not self.product in contexts:
            raise ContextException (f"undefined product {self.product} not found in contexts.")
        logger.info (f"-- load-context: id:{self.product}")
        '''
        context = contexts[self.product]
        apps = context.get ('apps', {})
        """ Resolve context inheritance. """
        for base_name in context.get ('extends', []):
            if not base_name in contexts:
                raise ContextException (f"base {base_name} of context {self.product} not found in registry.")
            logger.debug (f"resolving inheritance of base {base_name} by context {self.product}")
            apps.update (contexts[base_name].get('apps'))
            new_apps = contexts[base_name].get ('apps', {})
            new_apps.update (apps)
            apps = new_apps
        '''
        context = contexts[self.product]
        logger.debug (f"---------------> {context}")
        apps = self.inherit (contexts=contexts, context=context)
        self.mixin_defaults(apps)
        self.mixin(contexts,context,apps)
        """ Load the repository map to enable string interpolation. """
        repository_map = {
            key : value['url']
            for key, value in self.registry.get ('repositories', {}).items ()
        }
        """ Compile URLs to resolve repository variables. """
        for name, app in apps.items (): #context.get('apps',{}).items ():
            if not 'spec' in app:
                repos = list(repository_map.items())
                if len(repos) == 0:
                    raise ValueError ("No spec URL and no repositories specified.")
                repo_url = repos[0][1]
                dockstore_branch = os.environ.get("DOCKSTORE_APPS_BRANCH", "master")
                if dockstore_branch != "master":
                    repo_url = repo_url.replace("master", dockstore_branch)
                app['spec'] = f"{repo_url}/{name}/docker-compose.yaml"
            spec_url = app['spec']
            app['icon'] = os.path.join (os.path.dirname (spec_url), "icon.png")
            for key in [ 'spec', 'icon', 'docs' ]:
                url = app[key]
                app[key] = Template(url).safe_substitute (repository_map)
            logger.debug (f"-- spec: {app['spec']}")
            logger.debug (f"-- icon: {app['icon']}")
        logger.debug (f"-- product {self.product} resolution => apps: {apps.keys()}")
        apps = self.add_conf_impl(apps, context)
        for app, value in apps.items():
            logger.debug("app:" + app)
            logger.debug(value)
        return apps
    
    def get_definition(self, app_id):
        """ Get the apps source definition"""
        logger.debug(f"----> ENTER GET_DEFINITION . . .")
        app_definition = self.apps[app_id].get('definition')
        if not app_definition:
            try:
                logger.debug (f"-- resolving definition for {app_id}")
                url = self.apps[app_id]['spec']
                logger.debug (f"-----> SPEC URL: {url}")
                logger.debug (f"-----> FETCHING URL")
                response = self.http_session.get(url)
                # response is a str here . . .
                logger.debug(f"RESPONSE_TEXT:\n{response.text}\nTYPE={type(response.text)}")
                if response.status_code != 200:
                    raise ValueError(f"-- app {app_id}. failed to parse spec. code:{response.status_code}")
                # After yaml.safe_load, spec_template is a dict(!)
                template_dict = yaml.safe_load(response.text)
                logger.debug (f"-----> TEMPLATE_DICT:\n{template_dict}\nTYPE={type(template_dict)}")
                context = self.registry["settings"]
                logger.debug (f"-----> CONTEXT:\n{context}\nTYPE={type(context)}")
                # Convert spec_template to a string before passing into jinja2Template
                spec_template = yaml.safe_dump(template_dict)
                logger.debug (f"-----> SPEC_TEMPLATE:\n{spec_template}\nTYPE={type(spec_template)}")
                template = jinja2Template(spec_template)
                try:
                    app_definition = template.render(**context)
                    logger.debug (f"-----> APP_DEFINITION:\n{app_definition}")
                    self.apps[app_id]['definition'] = app_definition
                except Exception as e:
                    logger.error (f"-- app {app_id} failed to render app definition.\nError: {e}")
                    logger.debug("", exc_info=True)
                    logger.warning (f"-- Setting app {app_id} app definition to \{\}")
                    app_definition = {}
                    self.apps[app_id]['definition'] = app_definition
                    continue
            except Exception as e:
                logger.error (f"-- app {app_id} failed to parse definition.\nStatus code:{response.status_code}\nError: {e}")
                logger.debug("", exc_info=True)
        logger.debug(f"----> EXIT GET_DEFINITION . . .")
        return app_definition

    def get_spec (self, app_id):
        """ Get the URL of the system docker-compose yaml specification. """
        logger.debug(f"----> ENTER GET_SPEC . . .")
        spec = self.apps[app_id].get ('spec_obj', None)
        if not spec:
            url = None
            response = None
            try:
                logger.debug (f"-- resolving definition for {app_id}")
                url = self.apps[app_id]['spec']
                logger.debug (f"-----> SPEC URL: {url}")
                logger.debug (f"-----> FETCHING URL")
                response = self.http_session.get(url)
                # response is a str here . . .
                logger.debug(f"RESPONSE_TEXT:\n{response.text}\nTYPE={type(response.text)}")
                if response.status_code != 200:
                    raise ValueError(f"-- app {app_id}. failed to parse spec. code:{response.status_code}")
                # After yaml.safe_load, spec_template is a dict(!)
                template_dict = yaml.safe_load(response.text)
                logger.debug (f"-----> TEMPLATE_DICT: {template_dict}\nTYPE={type(template_dict)}")
                context = self.registry["settings"]
                logger.debug (f"-----> CONTEXT: {context}\nTYPE={type(context)}")
                # Convert spec_template to a string before passing into jinja2Template
                spec_template = yaml.safe_dump(template_dict)
                logger.debug (f"-----> SPEC_TEMPLATE: {spec_template}\nTYPE={type(spec_template)}")
                template = jinja2Template(spec_template)
                try:
                    spec = template.render(**context)
                    logger.debug (f"-----> RENDERED SPEC: {spec}")
                    self.apps[app_id]['spec_obj'] = spec
                except Exception as e:
                    traceback.print_exc()
                    logger.error (f"--  app {app_id}.\n Failed to render spec.\nError: {e}")
            except Exception as e:
                traceback.print_exc ()
                if response:
                    logger.error (f"-- app {app_id}. failed to parse spec. code:{response.status_code}")
                else:
                    logger.error (f"-- app {app_id}. failed to parse spec.")
                raise e
        else:
             try:
                 logger.debug(f"-- removing cached volumes")
                 spec['services'][app_id].pop('volumes', None)
             except Exception as e:
                 traceback.print_exc ()
                 logger.error (f"-- app {app_id}. failed to remove cached volumes in parse spec.")
                 raise e
        logger.debug(f"----> EXIT GET_SPEC")
        return spec

    def get_env_registry(self, app_id, settings):
        """ Get the env variables specified for an app in the registry and update settings"""
        env = self.apps[app_id].get('env', None)
        if env:
            settings.update(env)
        return settings
    
    def get_settings (self, app_id):
        """ Get the URL of the .env settings / environment file. """
        env = self.apps[app_id].get ('env_obj', None)
        if not env:
            url = self.apps[app_id]['spec']
            env_url = os.path.join (os.path.dirname (url), ".env")
            logger.debug (f"-- resolving settings for app: {app_id}")
            response = self.http_session.get (env_url)
            if response.status_code == 200:
                logger.debug (f"-- got settings for {app_id}")
                env = response.text
            else:
                logger.debug (f"-- using empty settings for {app_id}")
                env = ""
            self.apps[app_id]['env_obj'] = env
        return env

    def status (self, request):
        return self.client.status (request)

    def delete (self, request):
        return self.client.delete (request)

    def update(self, request):
        return self.client.patch(request)
    
    def start (self, principal, app_id, resource_request):
        """ Get application metadata, docker-compose structure, settings, and compose API request. """
        spec = self.get_spec (app_id)
        settings = self.client.parse_env (self.get_settings (app_id))
        settings_all = self.get_env_registry(app_id, settings)
        services = self.apps[app_id]['services']
        services = { k : {
            "port" : str(v),
            "clients" : []
          } for k, v in services.items ()
        }
        logger.debug (f"parsed {app_id} settings: {settings}")
        """ Use a pre-existing k8s service account """
        service_account = self.apps[app_id]['serviceAccount'] if 'serviceAccount' in self.apps[app_id].keys() else None
        """ Add entity's auth information """
        principal_params = {"username": principal.username, "access_token": principal.access_token, "refresh_token": principal.refresh_token}
        principal_params_json = json.dumps(principal_params, indent=4)
        """ Security Context that are set for the app """
        spec["services"][app_id]["securityContext"] = self.apps[app_id]["securityContext"] if 'securityContext' in self.apps[app_id].keys() else None
        spec["services"][app_id]["ext"] = self.apps[app_id]["ext"] if 'ext' in self.apps[app_id].keys() else None
        spec["services"][app_id].update(resource_request)
        """ Certain apps might require appending a string to the custom URL. """
        conn_string = self.apps.get(app_id).get("conn_string", "")
        spec["services"][app_id]["conn_string"] = conn_string
        """ Add a proxy rewrite rule """
        proxy_rewrite_rule = self.apps.get(app_id).get("proxy-rewrite-rule", False)
        spec["services"][app_id]["proxy_rewrite_rule"] = proxy_rewrite_rule

        if spec is not None:
            system = self._start ({
                "name"       : app_id,
                "serviceaccount": service_account,
                "env"        : settings_all,
                "system"     : spec,
                "principal"   : principal_params_json,
                "services"   : services
            })
            """ Validate resulting interfaces. """
            """
            TODO: 
              1. Check returned status.
              2. The Ambassador based URL removes the need to pass back a port. Confirm & delete port code.
            """
            running = { v.name : v.port for v in system.services }
            for name, port in services.items ():
                assert name in running, f"Svc {name} expected but {services.keys()} actually running."            
            logger.info (
                f"  -- started app id:{app_id} user:{principal.username} id:{system.identifier} services:{list(running.items ())}")
        return system
    
    def _start (self, request):
        """
        Control low level application launching (start) logic. 
        Also provides an anchor point to mock the service in unit tests.
        """
        return self.client.start (request)

class NullContext (TychoContext):
    """
    A null context to facilitate client development.
    """
    def __init__(self,product="common"):
        super ().__init__(product=product,stub=True)

    def status(self, request=None):
        """ Make up some rows. """
        identifier = uuid.uuid4 ()
        return TychoStatus (**{
            "status" : "success",
            "result" : [
                {
                    "name"          : f"jupyter-ds-{str(identifier)}",
                    "app_id"        : "jupyter-ds",
                    "sid"           : str(identifier),
                    "ip_address"    : 'x.y.z.m',
                    "port"          : "8080",
                    "creation_time" : "time"
                } for x in range(8000, 8005)
            ],
            "message" : "..."
        })

    def delete (self, request):
        """ Ingore deletes. """
        logger.debug (f"-- delete: {request}")
        
    def start (self, principal, app_id):
        logger.debug (f"-- start: {principal} {app_id}")        
        spec = self.get_spec (app_id)
        #settings = self.client.parse_env (self.get_settings (app_id))
        settings = self._parse_env (self.get_settings (app_id))
        services = self.apps[app_id]['services']
        return TychoSystem (**{
        "status" : "ok",
        "result" : {
            "name"       : self.apps[app_id]['name'],
            "sid"        : uuid.uuid4 (),
            "containers" :  {
                k : { 'ip_address' : 'x.y.z', 'port-1' : v }
                for k, v in services.items ()
            }
        },
        "message" : "mock: testing..."
    })
    
class ContextFactory:
    """ Flexible method for connecting to a TychoContext.
    Also, provide the null context for easy dev testing in appstore. """
    @staticmethod
    def get (product, context_type="null"):
        return {
            "null" : NullContext(product=product),
            "live" : TychoContext(product=product)
        }[context_type]
    
            
        
